#include <assert.h>

#include <stdio.h>
#include <string.h>
#ifdef	__BSD__
#include <stdlib.h>
#else
#include <malloc.h>
#endif	/* __BSD__ */

/* Known events: event->event_class */
typedef enum
{
	no_event,
	key_press_event,
	key_release_event,
	pointer_motion_event
}	event_class_t;

/* Known scancodes: event->scan_code */
enum
{
	key_unknown,
	key_escape,
	key_enter,
	key_up,
	key_left,
	key_right,
	key_down,
	key_left_mouse,
	key_middle_mouse,
	key_right_mouse
};

struct ui_event_s
{
	struct ui_event_s *next, *prev;
	event_class_t	event_class;
	int 	scan_code;
	int 	char_code;
	int 	pointer_x;
	int 	pointer_y;
}	*message_queue = NULL;

event_class_t
peek_next_event(int *scan_ret, int *char_ret, int *px_ret, int *py_ret)
{
	event_class_t retval;
	struct ui_event_s *ev;
	if (!message_queue)
		return no_event;
	/* remove tail */
	ev = message_queue->prev;
	if (ev->next == ev)
		message_queue = NULL;
	else {
		ev->prev->next = message_queue;
		message_queue->prev = ev->prev;
	}
	if (scan_ret)
		*scan_ret = ev->scan_code;
	if (char_ret)
		*char_ret = ev->char_code;
	if (px_ret)
		*px_ret = ev->pointer_x;
	if (py_ret)
		*py_ret = ev->pointer_y;
	retval = ev->event_class;
	free(ev);
	return retval;
}

void
push_event(event_class_t event_class, int scan_code, int char_code,
	   int pointer_x, int pointer_y)
{
	struct ui_event_s *ev;
	ev = (struct ui_event_s *)calloc(1, sizeof(struct ui_event_s));
	ev->event_class = event_class;
	ev->scan_code = scan_code;
	ev->char_code = char_code;
	ev->pointer_x = pointer_x;
	ev->pointer_y = pointer_y;
	if (!message_queue) {
		message_queue = ev->next = ev->prev = ev;
		return;
	}
	/* append to head */
	ev->prev = message_queue->prev;
	ev->next = message_queue;
	message_queue->prev->next = ev;
	message_queue->prev = ev;
	message_queue = ev;
}

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <GL/gl.h>
#include <GL/glx.h>
#ifdef	__BSD__
#include <unistd.h> /* for select() */
#else
#include <sys/select.h>
#endif	/* __BSD__ */
#include <sys/time.h>
#include <sys/types.h>

#define	HAVE_XF86_VIDMODE
/*
 * Uncomment to use XFree86 VidMode extension
 */

#ifdef	HAVE_XF86_VIDMODE
#include <X11/extensions/xf86vmode.h>
#endif	/* HAVE_XF86_VIDMODE */

Display 	*dpy;
int 		scrnum;
Window		root, wnd;
int 		wnd_focused = 0;
XSetWindowAttributes attr;

#ifdef	HAVE_XF86_VIDMODE

XF86VidModeModeInfo **vidmodes = NULL;
int num_vidmodes = 0;
int vidmode_changed = 0;

int
X11_set_video_mode(int width, int height)
{
	int mj = 0, mn = 0;
	int sw, sh, i;
	sw = WidthOfScreen(DefaultScreenOfDisplay(dpy));
	sh = HeightOfScreen(DefaultScreenOfDisplay(dpy));
	if (width == sw && height == sh)
		return !0;
	if (!vidmodes) {
		if (!XF86VidModeQueryVersion(dpy, &mj, &mn)) {
			fprintf(stderr, "XFree86-VidModeExtension not found.\n");
			return 0;
		}
		printf("found XFree86-VidModeExtension (version %d.%d).\n", mj, mn);
		XF86VidModeGetAllModeLines(dpy, scrnum, &num_vidmodes, &vidmodes);
		if (!vidmodes) { /* just a safety measure, though... */
			fprintf(stderr, "XF86VidModeGetAllModeLines() failed !.\n");
			return 0;
		}
	}
	for (i = 0; i < num_vidmodes; i++) {
		if (width == vidmodes[i]->hdisplay &&
			height == vidmodes[i]->vdisplay)
			break;
	}
	if (i == num_vidmodes) {
		fprintf(stderr, "Unable to find mode '%i X %i'.\n", width, height);
		return 0;
	}
	XWarpPointer(dpy, None, wnd, 0, 0, 0, 0, width >> 1, height >> 1);
	XF86VidModeSwitchToMode(dpy, scrnum, vidmodes[i]);
	XF86VidModeSetViewPort(dpy, scrnum, 0, 0);
	/* TODO: viewport locking (but _how_? - have to look to Q code ?) */
	vidmode_changed = 1;
	return !0;
}

void
X11_restore_video_mode(void)
{
	int i;
	if (!vidmodes)
		return;
	if (vidmode_changed) {
		XF86VidModeSwitchToMode(dpy, scrnum, vidmodes[0]);
		vidmode_changed = 0;
	}
	for (i = 0; i < num_vidmodes; i++) {
		if (vidmodes[0]->privsize)
#if	defined(__cplusplus) || defined(c_plusplus)
			XFree(vidmodes[0]->c_private);
#else
			XFree(vidmodes[0]->private);
#endif
	}
	XFree(vidmodes);
	vidmodes = NULL;
}

#else	/* !HAVE_XF86_VIDMODE */

int
X11_set_video_mode(int width, int height)
{
	return 0;
}

void
X11_restore_video_mode(void) 
{
}

#endif	/* !HAVE_XF86_VIDMODE */

/*
 * ref_shutdown()
 */
void
ref_shutdown(void)
{
	/* if (gl_drawable) {
		glXMakeCurrent(dpy, None, NULL);
		glXDestroyContext(dpy, gl_drawable);
		gl_drawable = NULL;
	} */
	if (wnd) {
		XDestroyWindow(dpy, wnd);
		wnd = 0;
	}
	X11_restore_video_mode();
	if (dpy) {
		XCloseDisplay(dpy);
		dpy = NULL;
	}
}

#include <stdlib.h>
#include <locale.h>

/*
 * ref_init()
 */
int
ref_init(char *caption, int width, int height, int fullscreen,
		 int need_depth, int need_stencil, int need_dst_alpha)
{
	int 	attrib[64], *aptr = attrib;
	unsigned long	mask;
	XSizeHints	sh;
	dpy = XOpenDisplay(NULL);
	if (!dpy) {
		fprintf(stderr, "XOpenDisplay() failed.\n");
		return 0;
	}
	scrnum = DefaultScreen(dpy);
	root = RootWindow(dpy, scrnum);
	*aptr++ = GLX_RGBA;
	*aptr++ = GLX_DOUBLEBUFFER;
	*aptr++ = GLX_RED_SIZE;
	*aptr++ = 1;
	*aptr++ = GLX_GREEN_SIZE;
	*aptr++ = 1;
	*aptr++ = GLX_BLUE_SIZE;
	*aptr++ = 1;
	*aptr = None;
	if (need_depth) {
		*aptr++ = GLX_DEPTH_SIZE;
		*aptr++ = 1;
		*aptr = None;
	}
	if (need_stencil) {
		*aptr++ = GLX_STENCIL_SIZE;
		*aptr++ = 1;
		*aptr = None;
	}
	if (need_dst_alpha)
		{
			*aptr++ = GLX_ALPHA_SIZE;
			*aptr++ = 1;
			*aptr = None;
		}

	{
		/* -> LibVCreateRoot() */
		char *lang = getenv("LANG");
		char *deflang = "en_US";
		if(!lang)
		{
			
			fprintf(stderr, "warning: LANG not set, assuming \"%s\".\n",deflang);
			lang = deflang;
		}
		if(!setlocale(LC_ALL,lang))
		{
			fprintf(stderr, "warning: could not set locale \"%s\"\n", lang);
		}
		else
			printf("Using locale \"%s\"\n", lang);
	}

	attr.border_pixel = None;
	attr.event_mask = StructureNotifyMask | ExposureMask
		| KeyPressMask | KeyReleaseMask | KeymapStateMask
		| ButtonPressMask | ButtonReleaseMask | PointerMotionMask
		| ButtonMotionMask | VisibilityChangeMask | FocusChangeMask
		/* | EnterWindowMask | LeaveWindowMask */;
	attr.do_not_propagate_mask = KeyPressMask | KeyReleaseMask |
		ButtonPressMask | ButtonReleaseMask | PointerMotionMask |
		ButtonMotionMask;
	attr.background_pixel = WhitePixel(dpy, scrnum);
	attr.backing_store = NotUseful;
	attr.save_under = False;
	mask = CWBorderPixel | CWEventMask | CWDontPropagate |
		CWBackPixel | CWBackingStore | CWSaveUnder;
	if (fullscreen) {
		mask |= CWOverrideRedirect;
		attr.override_redirect = True;
	}
	if (width == 0)
		width = DisplayWidth(dpy, scrnum);
	if (height == 0)
		height = DisplayHeight(dpy, scrnum);
	wnd = XCreateWindow(dpy, root, 0, 0, width, height, 0,
						DefaultDepth(dpy,scrnum), InputOutput,
						DefaultVisual(dpy,scrnum),
						mask, &attr);
	if (!wnd) {
		fprintf(stderr, "XCreateWindow() failed.\n");
		ref_shutdown();
		return 0;
	}
	XStoreName(dpy, wnd, caption);
	memset(&sh, 0, sizeof(sh));
	/*
	  sh.flags = USPosition | USSize | PPosition | PSize | \
	  PMinSize | PMaxSize | PBaseSize;
	*/
	sh.flags = USPosition | PBaseSize;
	sh.x = 0;
	sh.y = 0;
	sh.width = sh.min_width = sh.max_width = sh.base_width = width;
	sh.height = sh.min_height = sh.max_height = sh.base_height = height;
	XSetWMNormalHints(dpy, wnd, &sh);
	XMapRaised(dpy, wnd);
	if (fullscreen)
		XSetInputFocus(dpy, wnd, RevertToParent, CurrentTime);
	return !0;
}

int
X11_scan_code(XEvent *event, int *char_ret)
{
	char		buf[64];
	KeySym		keysym;
	/*
	 * FIXME
	 */
	if (XLookupString(&event->xkey, buf, sizeof(buf), &keysym, 0) == 1)
		*char_ret = buf[0];
	else
		*char_ret = 0;
	switch (keysym)
		{
		case XK_Escape:
			return key_escape;
		case XK_KP_Enter:
		case XK_Return:
			return key_enter;
		case XK_Left:
			return key_left;
		case XK_Right:
			return key_right;
		case XK_Down:
			return key_down;
		case XK_Up:
			return key_up;
		default:
			break;
		}
	return key_unknown;
}

int
X11_button_code(XEvent * event)
{
	switch (event->xbutton.button) {
	case Button1:return key_left_mouse;
	case Button2:
		return key_middle_mouse;
	case Button3:
		return key_right_mouse;
	default:
		break;
	}
	return key_unknown;
}

void
X11_grab_pointer(int grab)
{
	static int pointer_grabbed = 0;
	if (grab) {
		if (pointer_grabbed)
			return;
		XGrabPointer(dpy, wnd, True, 0,
					 GrabModeAsync, GrabModeAsync, wnd,
					 None, CurrentTime);
		pointer_grabbed = 1;
	} else {
		if (!pointer_grabbed)
			return;
		pointer_grabbed = 0;
		XUngrabPointer(dpy, CurrentTime);
	}
}

static void
dump (char *str, int len)
{
    printf("(");
    len--;
    while (len-- > 0)
        printf("%02x ", (unsigned char) *str++);
    printf("%02x)", (unsigned char) *str++);
}

static void
do_KeyPress (XEvent *eventp)
{
	XKeyEvent *e = (XKeyEvent *) eventp;
	KeySym ks;
	KeyCode kc = 0;
	Bool kc_set = False;
	char *ksname;
	int nbytes;
	char str[256+1];
	static char *buf = NULL;
	static int bsize = 8;

	if (buf == NULL)
		buf = malloc (bsize);

    nbytes = XLookupString (e, str, 256, &ks, NULL);
	if (ks == NoSymbol)
		ksname = "NoSymbol";
    else
	{
		ksname = XKeysymToString (ks);
		if (!ksname)
	    	ksname = "(no name)";
		kc = XKeysymToKeycode(dpy, ks);
		kc_set = True;
    }

    printf ("keycode %u (keysym 0x%lx, %s)\n",e->keycode,
		(unsigned long)ks, ksname);

    if (kc_set && e->keycode != kc)
		printf ("XKeysymToKeycode() returns keycode: %u\n",kc);
    if (nbytes < 0) nbytes = 0;
    if (nbytes > 256) nbytes = 256;
    str[nbytes] = '\0';
    printf ("    XLookupString gives %d bytes", nbytes);
    if (nbytes > 0)
	{
		printf(": ");
		dump (str, nbytes);
		printf (" \"%s\"\n", str);
	}
	else
	{
    	printf (".\n");
    }
}

void
X11_process_message(XEvent * event)
{
	int scan_code = 0, char_code = 0;
	event_class_t event_class;
	switch (event->type) {
	case FocusOut:
		wnd_focused = 0;
		puts("OUT");
		return;
	case FocusIn:
		wnd_focused = 1;
		puts("IN");
		return;
	case KeyPress:
		do_KeyPress(event);
		event_class = key_press_event;
		scan_code = X11_scan_code(event, &char_code);
		break;
	case KeyRelease:
		event_class = key_release_event;
		scan_code = X11_scan_code(event, &char_code);
		break;
	case MotionNotify:
		if (!wnd_focused)
			return;
		event_class = pointer_motion_event;
		scan_code = char_code = 0;
		break;
	case ConfigureNotify:
		printf("ConfigureNotify: (%d,%d) - [%d,%d]\n",
			event->xconfigure.x, event->xconfigure.y,
			event->xconfigure.width, event->xconfigure.height);
		return;
	case ButtonPress:
		event_class = key_press_event;
		scan_code = X11_button_code(event);
		char_code = 0;
	case ButtonRelease:
		event_class = key_release_event;
		scan_code = X11_button_code(event);
		char_code = 0;
		break;
	default:
/*		printf("UNKN %d\n",event->type); */
		return;
	}
	push_event(event_class, scan_code, char_code,
			   event->xmotion.x, event->xmotion.y);
}

event_class_t
ref_next_event(int *scan_ret, int *char_ret, int *px_ret, int *py_ret)
{
	XEvent event;
	while (XPending(dpy)) {
		XNextEvent(dpy, &event);
		X11_process_message(&event);
	}
	return peek_next_event(scan_ret, char_ret, px_ret, py_ret);
}

void											
ref_end_frame(void)
{
}

void
ref_delay(int ms)
{
	/*
	 * The straightforward way is to use usleep(), but there are some
	 * bizzare problems with headers (even) on modern Linux
	 * distrubutions.
	 * And: this approach also works under FreeBSD.
     * The idea is _not_ mine; actually, it's quite old...
	 */
	struct timeval	delay;
	delay.tv_sec = ms / 1000;
	delay.tv_usec = 1000 * (ms % 1000);
	select(0, NULL, NULL, NULL, &delay);
}

int main(void)
{
	event_class_t ev_cls;
	int code, chr, px, py;
	if (!ref_init("demo1", 640, 480, 0, 0, 0, 0))
		return -1;
	do {
		ev_cls = ref_next_event(&code, &chr, &px, &py);
		if (ev_cls == no_event) {
			ref_delay(20);
			continue;
		}
		/*
		printf("[%s]: CODE '%i', CHAR '%i' ('%c'), PX %i, PY %i\n",
			   classes[ev_cls], code, chr,
			   (chr >= 32) ? chr : ' ',
			   px, py);
		*/
	} while (code != key_escape);
	ref_shutdown();
	return 0;
}
